﻿using System;
using System.ComponentModel;

namespace LeakyBucketTest {
	class LeakyBucket {
		//---------------------------------------------------------------------------
		/// <summary> 最大リクエストクォータ </summary>
		//---------------------------------------------------------------------------
		public int MaxQuota    { set; get; }

		//---------------------------------------------------------------------------
		/// <summary> 回復レート(mSec) </summary>
		//---------------------------------------------------------------------------
		public int RestoreRate { set; get; }

		//---------------------------------------------------------------------------
		/// <summary> 現在値 </summary>
		//---------------------------------------------------------------------------
		private int _Value;
		public int Value { 
			set {
				lock( this ) {
					// 値が0だった時は、カウントダウンを今から始める
					if ( this._Value == 0 ) {
						lastCheckTime = DateTime.Now;
					}

					// 指定された値をセット(但し有効範囲は,0～MaxQuotaのみ)
					this._Value = Math.Min( MaxQuota, Math.Max( 0, value ) );
				}
			}
			get {
				return this._Value;
			}
		}

		//---------------------------------------------------------------------------
		// 最後に回復させた時刻
		//---------------------------------------------------------------------------
		private DateTime lastCheckTime = DateTime.Now;

		private BackgroundWorker backWorker = new BackgroundWorker();


		//---------------------------------------------------------------------------
		// コンストラクタ
		//---------------------------------------------------------------------------
		public LeakyBucket() {
			backWorker.DoWork          += backWorker_DoWork;
			backWorker.ProgressChanged += backWorker_ProgressChanged;
			backWorker.WorkerReportsProgress      = true;
			backWorker.WorkerSupportsCancellation = true;

			// 非同期処理を開始する
			backWorker.RunWorkerAsync();
		}

		//---------------------------------------------------------------------------
		// デストラクタ
		//---------------------------------------------------------------------------
		~LeakyBucket() {
			// 非同期処理をキャンセルする
			backWorker.CancelAsync();
		}

		//---------------------------------------------------------------------------
		// 非同期でのバックグラウンド処理のハンドラ
		//---------------------------------------------------------------------------
		private void backWorker_DoWork(object sender, DoWorkEventArgs e) {
			BackgroundWorker worker = (BackgroundWorker)sender;

			//--------------------------------------------
			// キャンセル要求があがるまで、ずっと繰り返し
			//--------------------------------------------
			while ( !worker.CancellationPending ) {
				System.Threading.Thread.Sleep( 20 );

				// 前回の回復時刻からの経過時間を求める
				TimeSpan restTime = DateTime.Now - lastCheckTime;

				// 指定時間が経過してなければ、何もしない
				if ( restTime.TotalMilliseconds < RestoreRate ) {
					continue;
				}

				lock( this ) {
					// 値を減算
					if ( Value > 0 ) {
						// メインスレッドに減算を依頼する
						worker.ReportProgress(0);
					}

					// 最終更新時刻を更新する
					lastCheckTime = DateTime.Now;
				}
			}
		}

		//---------------------------------------------------------------------------
		// 状態変更時のハンドラ
		//---------------------------------------------------------------------------
		private void backWorker_ProgressChanged( object sender, ProgressChangedEventArgs e ) {
			// 現在値を減算する
			lock( this ) {
				Value--;
			}
		}
	}
}
